# -*- coding: utf-8 -*-
# MIT License

# Copyright (c) 2020-present largecats

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import sys
import os
sys.path.insert(0, os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__)))))
import logging
from sparksqlformatter.src import api
from sparksqlformatter.src.formatter import Formatter
from sparksqlformatter.src.style import Style

logger = logging.getLogger(__name__)
log_formatter = '[%(asctime)s] %(levelname)s [%(filename)s:%(lineno)s:%(funcName)s] %(message)s'
logging.basicConfig(stream=sys.stdout, level=logging.INFO, format=log_formatter)


class Test:
    def __init__(self):
        pass

    def test_create_table_using_data_source(self):
        msg = 'Testing create table using data source'
        testQuery = '''
CREATE TABLE xxx
    (time string, amount double, country string, date date)
    USING org.apache.spark.sql.parquet
    OPTIONS ('serialization.format'='1', 'path'='/user/xxx', 'mergeSchema'='true')
    PARTITIONED BY (country, date)
        '''
        key = '''
CREATE TABLE
    xxx (time string, amount double, country string, date date)
    USING org.apache.spark.sql.parquet
    OPTIONS ('serialization.format' = '1', 'path' = '/user/xxx', 'mergeSchema' = 'true')
    PARTITIONED BY (country, date)
        '''.strip()
        return self.run(msg, testQuery, key)

    def test_insert_without_into(self):
        msg = 'Testing INSERT without INTO'
        testQuery = '''INSERT Customers (ID, MoneyBalance, Address, City) VALUES (12, -123.4, 'Skagen 2111','Stv')'''
        key = '''
INSERT
    Customers (ID, MoneyBalance, Address, City)
VALUES
    (12, -123.4, 'Skagen 2111', 'Stv')
        '''.strip()
        return self.run(msg, testQuery, key)

    def test_alter_table_modify(self):
        msg = 'Testing ALTER TABLE ... MODIFY query'
        testQuery = '''ALTER TABLE supplier MODIFY supplier_name STRING(100) NOT NULL'''
        key = '''
ALTER TABLE
    supplier
MODIFY
    supplier_name STRING(100) NOT NULL
        '''.strip()
        return self.run(msg, testQuery, key)

    def test_alter_table_alter_column(self):
        msg = 'Testing ALTER TABLE ... ALTER COLUMN query'
        testQuery = '''ALTER TABLE supplier ALTER COLUMN supplier_name STRING(100) NOT NULL'''
        key = '''
ALTER TABLE
    supplier
ALTER COLUMN
    supplier_name STRING(100) NOT NULL
        '''.strip()
        return self.run(msg, testQuery, key)

    def test_simple_select(self):
        msg = 'Testing simple SELECT'
        testQuery = '''SELECT c1, c2 FROM t0'''
        key = '''
SELECT
    c1,
    c2
FROM
    t0
        '''.strip()
        return self.run(msg, testQuery, key)

    def test_select_with_left_join(self):
        msg = 'Testing SELECT query with LEFT JOIN'
        testQuery = '''SELECT a, b FROM t LEFT JOIN t2 ON t.id = t2.id_t'''
        key = '''
SELECT
    a,
    b
FROM
    t
LEFT JOIN
    t2
    ON t.id = t2.id_t
        '''.strip()
        return self.run(msg, testQuery, key)

    def test_select_with_function(self):
        msg = 'Testing SELECT with udf'
        testQuery = '''SELECT from_unixtime(time), c2 FROM t0'''
        key = '''
SELECT
    from_unixtime(time),
    c2
FROM
    t0
        '''.strip()
        return self.run(msg, testQuery, key)

    def test_select_with_udf(self):
        msg = 'Testing SELECT with udf'
        testQuery = '''SELECT foo(c1), c2 FROM t0'''
        key = '''
SELECT
    foo(c1),
    c2
FROM
    t0
        '''.strip()
        return self.run(msg, testQuery, key, {'userDefinedFunctions': ['foo']})

    def test_case_when(self):
        msg = 'Testing CASE ... WHEN'
        testQuery = '''SELECT c1, c2, CASE WHEN c3 = 'one' THEN 1 WHEN c3 = 'two' THEN 2 ELSE 3 END AS c4 FROM t0'''
        key = '''
SELECT
    c1,
    c2,
    CASE
        WHEN c3 = 'one'
        THEN 1
        WHEN c3 = 'two'
        THEN 2
        ELSE 3
    END AS c4
FROM
    t0
        '''.strip()
        return self.run(msg, testQuery, key)

    def test_case_when_with_expression(self):
        msg = 'Testing CASE ... WHEN with expression'
        testQuery = '''
SELECT
CASE WHEN toString(getNumber()) = 'one' THEN 1
WHEN toString(getNumber()) = 'two' THEN 2 ELSE 3 END AS c1
FROM t0
'''
        key = '''
SELECT
    CASE
        WHEN toString(getNumber()) = 'one'
        THEN 1
        WHEN toString(getNumber()) = 'two'
        THEN 2
        ELSE 3
    END AS c1
FROM
    t0
        '''.strip()
        return self.run(msg, testQuery, key)

    def test_lowercase_case_when(self):
        msg = 'Testing lower-case CASE ... WHEN'
        testQuery = '''select case when c1 = 'foo' then 1 else 2 end from t0'''
        key = '''
SELECT
    CASE
        WHEN c1 = 'foo'
        THEN 1
        ELSE 2
    END
FROM
    t0
        '''.strip()
        return self.run(msg, testQuery, key)

    def test_ignore_case_when_in_other_string(self):
        msg = 'Testing ignore CASE, WHEN in other strings'
        testQuery = '''SELECT CASEDATE, ENDDATE FROM table1'''
        key = '''
SELECT
    CASEDATE,
    ENDDATE
FROM
    table1
        '''.strip()
        return self.run(msg, testQuery, key)

    def test_nested_case_when(self):
        msg = 'Testing nested CASE ... WHEN'
        testQuery = '''
select
    case when a > 0 then
        case when b > 0 then True
        else False
        end
    end
    from t0
        '''
        key = '''
SELECT
    CASE
        WHEN a > 0
        THEN CASE
            WHEN b > 0
            THEN TRUE
            ELSE FALSE
        END
    END
FROM
    t0
        '''.strip()
        return self.run(msg, testQuery, key)

    def test_operator_spacing(self):
        msg = 'Testing operator spacing, e.g., -, &, {, }'
        testQuery = '''
select
    *,
    case when (a>0) and (cast(b as int) & {KEYWORD}=0)
    then -c
    else a-c end
    from t0
    where id != 1 or amount >= 0
        '''
        key = '''
SELECT
    *,
    CASE
        WHEN (a > 0) AND (CAST(b AS int) & {KEYWORD} = 0)
        THEN -c
        ELSE a - c
    END
FROM
    t0
WHERE
    id != 1
    OR amount >= 0
        '''.strip()
        return self.run(msg, testQuery, key)

    def test_line_comment(self):
        msg = 'Testing line comment'
        testQuery = '''
SELECT a,--comment, here
b FROM t0--comment'''
        key = '''
SELECT
    a, --comment, here
    b
FROM
    t0 --comment
        '''.strip()
        return self.run(msg, testQuery, key)

    def test_line_comment_followed_by_comma(self):
        msg = 'Testing line comment followed by comma'
        testQuery = '''
SELECT a --comment
, b
        '''
        key = '''
SELECT
    a --comment
,
    b
        '''.strip()
        return self.run(msg, testQuery, key)

    def test_parentheses(self):
        msg = 'Testing parentheses'
        testQuery = '''
SELECT COALESCE(collect_list(time)[0], 0) AS time
from t0
        '''
        key = '''
SELECT
    COALESCE(collect_list(time)[0], 0) AS time
FROM
    t0
        '''.strip()
        return self.run(msg, testQuery, key)

    def test_line_comment_followed_by_close_paren(self):
        msg = 'Testing line comment followed by closing parentheses'
        testQuery = '''
SELECT ( a --comment
 )
        '''
        key = '''
SELECT
    (
        a --comment
    )
        '''.strip()
        return self.run(msg, testQuery, key)

    def test_line_comment_followed_by_open_paren(self):
        msg = 'Testing line comment followed by opening parentheses'
        testQuery = '''
SELECT a --comment
()
        '''
        key = '''
SELECT
    a --comment
    ()
        '''.strip()
        return self.run(msg, testQuery, key)

    def test_block_comment(self):
        msg = 'Testing block comment'
        testQuery = '''
select
    a,
    /* case
        when a > 0 then true
        else false
    end as is_positive, */
    b
from t0
        '''
        key = '''
SELECT
    a,
    /* case
    when a > 0 then true
    else false
    end as is_positive, */
    b
FROM
    t0
        '''.strip()
        return self.run(msg, testQuery, key)

    def test_query_with_between_and(self):
        msg = 'Testing query with BETWEEN ... AND'
        testQuery = '''
select c1, c2 from t0
where c1 between '{date}' and add_months('{date}', 1)
and c2 < 0
        '''
        key = '''
SELECT
    c1,
    c2
FROM
    t0
WHERE
    c1 BETWEEN '{date}' AND add_months('{date}', 1)
    AND c2 < 0
        '''.strip()
        return self.run(msg, testQuery, key)

    def test_query_with_left_join_and(self):
        msg = 'Testing query with LEFT JOIN ... AND'
        testQuery = '''
select c1, c2 from t0
left join t1 on t0.c1 = t1.c1 and t0.c3 = t1.c3
        '''
        key = '''
SELECT
    c1,
    c2
FROM
    t0
LEFT JOIN
    t1
    ON t0.c1 = t1.c1
    AND t0.c3 = t1.c3
        '''.strip()
        return self.run(msg, testQuery, key)

    def test_query_with_nested_and_in_where(self):
        msg = 'Testing query with nested AND in WHERE'
        testQuery = '''
SELECT c1, c2 FROM t0 WHERE type = 1 OR (c3 = 0 AND type = 2)
        '''
        key = '''
SELECT
    c1,
    c2
FROM
    t0
WHERE
    type = 1
    OR (
        c3 = 0
        AND type = 2
    )
        '''.strip()
        return self.run(msg, testQuery, key)

    def test_query_with_over_clause(self):
        msg = 'Testing query with OVER clause'
        testQuery = '''
SELECT 
    *, 
    ROW_NUMBER() OVER(PARTITION BY a, b ORDER BY b desc, c desc) AS rank
FROM 
    t0
        '''
        key = '''
SELECT
    *,
    ROW_NUMBER() OVER(
        PARTITION BY
            a,
            b
        ORDER BY
            b DESC,
            c DESC
    ) AS rank
FROM
    t0
        '''.strip()
        return self.run(msg, testQuery, key)

    def test_query_with_group_by(self):
        msg = 'Testing query with GROUP BY'
        testQuery = '''
select * from t0 group by 1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,26,27,28,29,30,31,32,33,34,35,36,37,38,39,40
        '''
        key = '''
SELECT
    *
FROM
    t0
GROUP BY
    1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31,
    32, 33, 34, 35, 36, 37, 38, 39, 40
        '''.strip()
        return self.run(msg, testQuery, key, {'splitOnComma': False})

    def test_query_with_order_by_with_desc(self):
        msg = 'Testing query with ORDER BY with DESC'
        testQuery = '''
select * from t0 order by 1 desc, 2 desc, 3 desc, 4 desc, 5 desc, 6 asc, 7 asc, 8 desc, 9, 10, 11 desc, 12 desc, 13 desc, 14 desc, 15 desc, 16 desc, 
17 desc
        '''
        key = '''
SELECT
    *
FROM
    t0
ORDER BY
    1 DESC, 2 DESC, 3 DESC, 4 DESC, 5 DESC, 6 ASC, 7 ASC, 8 DESC, 9, 10, 11 DESC, 12 DESC, 13 DESC, 14 DESC, 15 DESC,
    16 DESC, 17 DESC
        '''.strip()
        return self.run(msg, testQuery, key, {'splitOnComma': False})

    def test_query_with_group_by_split_on_comma(self):
        msg = 'Testing query with GROUP BY'
        testQuery = '''
select * from t0 group by 1,2,3,4
        '''
        key = '''
SELECT
    *
FROM
    t0
GROUP BY
    1,
    2,
    3,
    4
        '''.strip()
        return self.run(msg, testQuery, key, {'splitOnComma': True})

    def test_query_with_order_by(self):
        msg = 'Testing query with ORDER BY'
        testQuery = '''
select * from t0 order by 1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,26,27,28,29,30,31,32,33,34,35,36,37,38,39,40
        '''
        key = '''
SELECT
    *
FROM
    t0
ORDER BY
    1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31,
    32, 33, 34, 35, 36, 37, 38, 39, 40
        '''.strip()
        return self.run(msg, testQuery, key, {'splitOnComma': False})

    def test_query_with_long_table_name(self):
        msg = 'Testing query with long table name'
        testQuery = '''
select t0.a, t0.b, t1.x, t2.y,
    t3.c, t3.d

from t0_very_very_very_very_very_very_very_very_very_long_table_name t0
    left join t1_very_very_very_very_very_very_very_very_very_long_table_name t1 on t0.a = t1.z
    left join t2 on t0.a = t2.z

    left join t3 on t3.c = t0.a

where t0.a between '{date}' and add_months('{date}', 1)
and t2.y < 0
order by 1,2,3
        '''
        key = '''
SELECT
    t0.a,
    t0.b,
    t1.x,
    t2.y,
    t3.c,
    t3.d
FROM
    t0_very_very_very_very_very_very_very_very_very_long_table_name t0
LEFT JOIN
    t1_very_very_very_very_very_very_very_very_very_long_table_name t1
    ON t0.a = t1.z
LEFT JOIN
    t2
    ON t0.a = t2.z
LEFT JOIN
    t3
    ON t3.c = t0.a
WHERE
    t0.a BETWEEN '{date}' AND add_months('{date}', 1)
    AND t2.y < 0
ORDER BY
    1,
    2,
    3
        '''.strip()
        return self.run(msg, testQuery, key)

    def test_query_with_unicode_char(self):
        msg = 'Testing query with unicode character'
        testQuery = '''
select c1, c2 from t0
where t1.c2 IN ('你好' 'ไหว้')
        '''
        key = '''
SELECT
    c1,
    c2
FROM
    t0
WHERE
    t1.c2 IN ('你好' 'ไหว้')
        '''.strip()
        return self.run(msg, testQuery, key)

    def test_query_with_string_formatting_with_keyword(self):
        msg = 'Testing query with mutliple {} for Python string formatting that contain keywords in the query'
        testQuery = '''
select * from t0 where t0.id in ({CREATE}, {UPDATE}, {UPDATE}, {UPDATE}, {UPDATE}, {UPDATE}, {UPDATE}, {UPDATE}, {UPDATE}, {UPDATE}, {UPDATE}, {UPDATE}, {UPDATE}) --CREATE or UPDATE
        '''
        key = '''
SELECT
    *
FROM
    t0
WHERE
    t0.id IN (
        {CREATE},
        {UPDATE},
        {UPDATE},
        {UPDATE},
        {UPDATE},
        {UPDATE},
        {UPDATE},
        {UPDATE},
        {UPDATE},
        {UPDATE},
        {UPDATE},
        {UPDATE},
        {UPDATE}
    ) --CREATE or UPDATE
        '''.strip()
        return self.run(msg, testQuery, key)

    def test_query_with_subquery(self):
        msg = 'Testing query with subquery'
        testQuery = '''
with t0 as (select c1, c2 from tab1),

t1 as (select c1, c2 from tab2),

t2 as (select c1, c2 from tab3)

select * from t0 left join t1 on t0.c1 = t1.c1
left join t2 on t0.c1 = t2.c1
        '''
        key = '''
WITH t0 AS (
    SELECT
        c1,
        c2
    FROM
        tab1
),

t1 AS (
    SELECT
        c1,
        c2
    FROM
        tab2
),

t2 AS (
    SELECT
        c1,
        c2
    FROM
        tab3
)

SELECT
    *
FROM
    t0
LEFT JOIN
    t1
    ON t0.c1 = t1.c1
LEFT JOIN
    t2
    ON t0.c1 = t2.c1
        '''.strip()
        return self.run(msg, testQuery, key)

    def test_query_with_subquery_inline_comment(self):
        msg = 'Testing query with subquery and inline comment'
        testQuery = '''
with t0 as -- comment 
(
    select * from t1
)

select * from t0
        '''
        key = '''
WITH t0 AS -- comment
(
    SELECT
        *
    FROM
        t1
)

SELECT
    *
FROM
    t0
        '''.strip()
        return self.run(msg, testQuery, key)

    def test_query_with_nested_subquery(self):
        msg = 'Testing query with nested subquery'
        testQuery = '''
select
    *,
    from_unixtime(ctime)
from
    (
        select * from t1
    )
        '''
        key = '''
SELECT
    *,
    from_unixtime(ctime)
FROM
    (
        SELECT
            *
        FROM
            t1
    )
        '''.strip()
        return self.run(msg, testQuery, key)

    def test_query_with_lateral_view_explode(self):
        msg = 'Testing query with LATERAL VIEW EXPLODE'
        testQuery = '''
select * from
    t0
    LATERAL VIEW EXPLODE(t0.groupA.list) t as groupA_list_explode
        '''
        key = '''
SELECT
    *
FROM
    t0
    LATERAL VIEW EXPLODE(t0.groupA.list) t AS groupA_list_explode
        '''.strip()
        return self.run(msg, testQuery, key)

    def test_reservedKeywordUppercase_style(self):
        msg = 'Testing reservedKeywordUppercase style'
        testQuery = '''select c1, c2 from t0'''
        key = '''
select
    c1,
    c2
from
    t0
        '''.strip()
        return self.run(msg, testQuery, key, {'reservedKeywordUppercase': False})

    def test_linesBetweenQueries_style(self):
        msg = 'Testing linesBetweenQueries style'
        testQuery = '''
with t0 as (select c1, c2 from tab1),

t1 as (select c1, c2 from tab2),

t2 as (select c1, c2 from tab3)

select * from t0 left join t1 on t0.c1 = t1.c1
left join t2 on t0.c1 = t2.c1
        '''
        key = '''
WITH t0 AS (
    SELECT
        c1,
        c2
    FROM
        tab1
),


t1 AS (
    SELECT
        c1,
        c2
    FROM
        tab2
),


t2 AS (
    SELECT
        c1,
        c2
    FROM
        tab3
)


SELECT
    *
FROM
    t0
LEFT JOIN
    t1
    ON t0.c1 = t1.c1
LEFT JOIN
    t2
    ON t0.c1 = t2.c1
        '''.strip()
        return self.run(msg, testQuery, key, {'linesBetweenQueries': 2})

    def run(self, msg, testQuery, key, style=Style()):
        logger.info(msg)
        logger.info('testQuery =')
        logger.info(testQuery)
        formattedQuery = api.format_query(testQuery, style)
        logger.info('formattedQuery =')
        logger.info(formattedQuery)
        logger.info(repr(formattedQuery))
        logger.info('key =')
        logger.info(key)
        logger.info(repr(key))
        assert formattedQuery == key
        return True

    def run_all(self):
        tests = list(filter(lambda m: m.startswith('test_'), dir(self)))
        for test in tests:
            getattr(self, test)()


if __name__ == "__main__":
    Test().run_all()



# -*- coding: utf-8 -*-
# MIT License

# Copyright (c) 2020-present largecats

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from sparksqlformatter.src.tokenizer import TokenType


class SubQuery:
    '''
    Class for maintaining the start and end of sub-queries.

    A sub-query is a query starting with
    WITH t0 AS (
        ...
    )
    or simply
    ...,
    t0 AS (
        ...
    )

    It needs to be succeeded by a comma if and only if the following statement is another sub-query. E.g.:

    WITH t0 AS (
        ...
    ),

    t1 AS (
        ...
    )

    SELECT * FROM t0
    '''
    def __init__(self):
        self.stack = []  # stack to hold unmatched opening and closing parentheses
        self.started = False  # whether a subquery has started, e.g., t0 AS (...)

    def matched(self):
        '''
        Check if all parentheses in the subquery have matched.

        Return: bool
        '''
        return len(self.stack) == 0

    def ended(self):
        '''
        Check if the subquery has ended.

        Return: bool
        '''
        return self.started and len(self.stack) == 0  # subQuery has started and parentheses are matched

    def update(self, formatter, token):
        '''
        Update self.stack and self.started with current token.

        Parameters
        formatter: sparksqlformatter.src.formatter.Formatter() object
            The formatter currently in use.
        token: sparksqlformatter.src.tokenizer.Token() object
            The current token.
        '''
        openParens = [p for p in formatter.style.openParens if p != 'CASE']
        closeParens = [p for p in formatter.style.closeParens if p != 'END']
        if token.value in openParens:  # push opening parenthesis onto stack
            self.stack.append(token.value)
        elif token.value in closeParens:
            if self.matched():  # no opening parenthesis to match
                raise Exception('Parentheses not matched')
            lastParen = self.stack[-1]  # retrieve last opening parenthesis
            if openParens.index(lastParen) == closeParens.index(token.value):  # check if they match
                self.stack.pop()  # remove the matched opening parenthesis
            else:
                raise Exception('Parentheses not matched')

    def reset(self):
        '''
        Reset the stack and started attributes of the class.
        '''
        self.started = False



# -*- coding: utf-8 -*-
# MIT License

# Copyright (c) 2016-present ZeroTurnaround LLC
# Copyright (c) 2016-present kufii
# Copyright (c) 2020-present largecats

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from __future__ import print_function  # for print() in Python 2
import re

import sparksqlformatter.src.config as config


class TokenType:
    '''
    Different types of tokens.
    '''
    WHITESPACE = 'WHITESPACE'
    WORD = 'WORD'
    KEYWORD = 'KEYWORD'
    STRING = 'STRING'
    RESERVED_KEYWORD = 'RESERVED_KEYWORD'
    TOP_LEVEL_KEYWORD = 'TOP_LEVEL_KEYWORD'
    TOP_LEVEL_KEYWORD_NO_INDENT = 'TOP_LEVEL_KEYWORD_NO_INDENT'
    NEWLINE_KEYWORD = 'NEWLINE_KEYWORD'
    OPERATOR = 'OPERATOR'
    OPEN_PAREN = 'OPEN_PAREN'
    CLOSE_PAREN = 'CLOSE_PAREN'
    LINE_COMMENT = 'LINE_COMMENT'
    BLOCK_COMMENT = 'BLOCK_COMMENT'
    NUMBER = 'NUMBER'


class Token:
    '''
    A token is a string that forms a unit in formatting.
    '''
    def __init__(self, type, value, flag=None):
        __slots__ = 'type', 'value', 'flag'  # saves space since there would be many instances of Token
        self.type = type
        self.value = value
        self.flag = flag  # added by formatter.py when formatting


class Tokenizer:
    def __init__(self, style):
        self.WHITESPACE_REGEX = u'^(\s+)'
        self.NUMBER_REGEX = r'^((-\s*)?[0-9]+(\.[0-9]+)?|0x[0-9a-fA-F]+|0b[01]+)\b'
        self.OPERATOR_REGEX = u'^([^\{\}]!=|<>|==|<=|>=|!=|!<|!>|\|\||::|->>|->|~~\*|~~|!~~\*|!~~|~\*|!~\*|!~|:=|.)'

        self.BLOCK_COMMENT_REGEX = u'(\/\*(?s).*?\*\/)'  # (?s) is inline flag for re.DOTALL
        self.LINE_COMMENT_REGEX = Tokenizer.create_line_comment_regex(style.lineCommentTypes)

        self.TOP_LEVEL_KEYWORD_REGEX = Tokenizer.create_keyword_regex(style.topLevelKeywords)
        self.TOP_LEVEL_KEYWORD_NO_INDENT_REGEX = Tokenizer.create_keyword_regex(style.topLevelKeywordsNoIndent)
        self.NEWLINE_KEYWORD_REGEX = Tokenizer.create_keyword_regex(style.newlineKeywords)
        self.RESERVED_KEYWORD_REGEX = Tokenizer.create_keyword_regex(config.Keyword.RESERVED_KEYWORDS)
        self.PLAIN_KEYWORD_REGEX = Tokenizer.create_keyword_regex(
            config.Keyword.RESERVED_KEYWORDS + config.Keyword.NON_RESERVED_KEYWORDS +
            config.Function.AGGREGATE_FUNCTIONS + config.Function.ARRAY_FUNCTIONS +
            config.Function.CONDITIONAL_FUNCTIONS + config.Function.DATE_TIME_FUNCTIONS +
            config.Function.HASH_FUNCTIONS + config.Function.JSON_FUNCTIONS + config.Function.MAP_FUNCTIONS +
            config.Function.MATHEMATICAL_FUNCTIONS + config.Function.MISC_FUNCTIONS +
            config.Function.OPERATOR_FUNCTIONS + config.Function.STRING_FUNCTIONS + config.Function.STRUCT_FUNCTIONS +
            config.Function.TABLE_GENERATING_FUNCTIONS + config.Function.TYPE_CONVERSION_FUNCTIONS +
            config.Function.WINDOW_FUNCTIONS + config.Function.XPATH_FUNCTIONS + style.userDefinedFunctions)

        self.WORD_REGEX = Tokenizer.create_word_regex(style.specialWordChars)
        self.STRING_REGEX = Tokenizer.create_string_regex(style.stringTypes)

        self.OPEN_PAREN_REGEX = Tokenizer.create_paren_regex(style.openParens)
        self.CLOSE_PAREN_REGEX = Tokenizer.create_paren_regex(style.closeParens)

    @staticmethod
    def create_line_comment_regex(lineCommentTypes):
        '''
        Create regex pattern that matches line comments in the query.

        Parameters
        lineCommentTypes: list
            Attribute of hiveqlformatter.src.style.Style() object.
        
        Return: string
            Regex pattern that matches line comment.
        '''
        lineCommentTypesString = ('|').join(list(map(lambda c: re.escape(c), lineCommentTypes)))
        regexString = u'^((?:{lineCommentTypesString}).*?(?:\r\n|\r|\n|$))'.format(
            lineCommentTypesString=lineCommentTypesString)
        return regexString

    @staticmethod
    def create_keyword_regex(keywords):
        '''
        Create regex pattern that matches keywords ikn the query.

        Parameters
        keywords: list
            Attribute of hiveqlformatter.src.style.Style() object.
        
        Return: string
            Regex pattern that matches keywords.
        '''
        keywordsString = ('|').join(keywords)
        keywordsPattern = re.sub(
            pattern=' ', repl='\\\s+', string=keywordsString
        )  # https://stackoverflow.com/questions/58328587/python-3-7-4-re-error-bad-escape-s-at-position-0
        regexString = u'^({keywordsPattern})\\b'.format(keywordsPattern=keywordsPattern)
        return regexString

    @staticmethod
    def create_word_regex(specialChars=[]):
        '''
        Create regex pattern that matches special characters in the query language.

        Parameters
        specialChars: list
            Attribute of hiveqlformatter.src.style.Style() object.
        
        Return: string
            Regex pattern that matches word with special characters.     
        '''
        specialCharsString = ('|').join(specialChars)
        regexString = u'^([\\w{specialCharsString}]+)'.format(specialCharsString=specialCharsString)
        return regexString

    @staticmethod
    def create_string_regex(stringTypes):
        '''
        Create regex pattern that matches strings in the query.

        Parameters
        stringTypes: list
            Attribute of hiveqlformatter.src.style.Style() object.
        
        Return: string
            Regex pattern that matches strings.    
        '''
        regexString = '^(' + Tokenizer.create_string_pattern(stringTypes) + ')'
        return regexString

    @staticmethod
    def create_string_pattern(stringTypes):
        '''This enables the following string patterns:
        1. curly bracket quoted python formatting keyword to be treated as string in hiveql formatting
        2. double quoted string using "" or \" to escape
        3. single quoted string using '' or \' to escape

        Parameters
        stringTypes: list
            Attribute of hiveqlformatter.src.style.Style() object.
        
        Return: string
            Regex pattern that matches strings.    
        '''
        patterns = {
            '{}': '(({[^}\\\\]*(?:\\\\.[^{\\\\]*)*(}|$))+)',
            '""': '(("[^"\\\\]*(?:\\\\.[^"\\\\]*)*("|$))+)',
            "''": "(('[^'\\\\]*(?:\\\\.[^'\\\\]*)*('|$))+)"
        }
        return ('|').join(list(map(lambda t: patterns[t], stringTypes)))

    @staticmethod
    def create_paren_regex(parens):
        '''
        Create regex pattern that matches parentheses in the query.

        Parameters
        parens: list
            Attribute of hiveqlformatter.src.style.Style() object.
        
        Return: string
            Regex pattern that matches parentheses.
        '''
        parensString = ('|').join(list(map(lambda p: Tokenizer.escape_paren(p), parens)))
        return '^(' + parensString + ')'

    @staticmethod
    def escape_paren(paren):
        '''
        Escape given parenthesis in regex pattern.

        Parameters
        paren: string
            The parenthesis to escape.
        
        Return: string
            Regex pattern with the parenthesis escaped.
        '''
        if (len(paren) == 1):
            return re.escape(paren)
        else:
            return '\\b' + paren + '\\b'

    def tokenize(self, input):
        """
        Takes a SQL string and breaks it into tokens.
        Each token is an object with type and value.

        Parameters
        input: string
            The query to format.
        
        Return: list
            Tokens extracted from the query.
        """
        if not input:
            return []

        tokens = []
        token = None
        while len(input):
            # Keep processing the string until it is empty
            token = self.get_next_token(input, token)  # get next token
            start = 0 if token is None else len(token.value)
            input = input[start::]  # advance thte string
            tokens.append(token)

        return tokens

    def get_next_token(self, input, previousToken):
        '''
        Find the next token object in given query.

        Parameters
        input: string
            The query to format.
        previousToken: Token() object
            The previous token found in the query.
        
        Return: Token() object
            The next token in the query.
        '''
        return (self.get_white_space_token(input) or self.get_comment_token(input) or self.get_string_token(input)
                or self.get_open_paren_token(input) or self.get_close_paren_token(input) or self.get_number_token(input)
                or self.get_keyword_token(input, previousToken) or self.get_word_token(input)
                or self.get_operator_token(input))

    def get_white_space_token(self, input):
        '''
        Find token of type TokenType.WHITESPACE in the query.

        Parameters
        input: string
            The query to format.
        
        Return: Token() object
            The identified white space token.
        '''
        return Tokenizer.get_token_on_first_match(input=input, type=TokenType.WHITESPACE, regex=self.WHITESPACE_REGEX)

    def get_comment_token(self, input):
        '''
        Find comment in the query.

        Parameters
        input: string
            The query to format.
        
        Return: Token() object
            The identified comment token.
        '''
        return self.get_line_comment_token(input) or self.get_block_comment_token(input)

    def get_line_comment_token(self, input):
        '''
        Find token of type TokenType.LINE_COMMENT in the query.

        Parameters
        input: string
            The query to format.
        
        Return: Token() object
            The identified line comment token.
        '''
        return Tokenizer.get_token_on_first_match(input=input,
                                                  type=TokenType.LINE_COMMENT,
                                                  regex=self.LINE_COMMENT_REGEX)

    def get_block_comment_token(self, input):
        '''
        Find token of type TokenType.BLOCK_COMMENT in the query.

        Parameters
        input: string
            The query to format.
        
        Return: Token() object
            The identified block comment token.
        '''
        return Tokenizer.get_token_on_first_match(input=input,
                                                  type=TokenType.BLOCK_COMMENT,
                                                  regex=self.BLOCK_COMMENT_REGEX)

    def get_string_token(self, input):
        '''
        Find token of type TokenType.STRING in the query.

        Parameters
        input: string
            The query to format.
        
        Return: Token() object
            The identified string token.
        '''
        return Tokenizer.get_token_on_first_match(input=input, type=TokenType.STRING, regex=self.STRING_REGEX)

    def get_open_paren_token(self, input):
        '''
        Find token of type TokenType.OPEN_PAREN in the query.

        Parameters
        input: string
            The query to format.
        
        Return: Token() object
            The identified opening parenthesis token.
        '''
        return Tokenizer.get_token_on_first_match(input=input, type=TokenType.OPEN_PAREN, regex=self.OPEN_PAREN_REGEX)

    def get_close_paren_token(self, input):
        '''
        Find token of type TokenType.CLOSE_PAREN in the query.

        Parameters
        input: string
            The query to format.
        
        Return: Token() object
            The identified closing parenthesis token.
        '''
        return Tokenizer.get_token_on_first_match(input=input, type=TokenType.CLOSE_PAREN, regex=self.CLOSE_PAREN_REGEX)

    def get_number_token(self, input):
        '''
        Find token of type TokenType.NUMBER in the query.

        Parameters
        input: string
            The query to format.
        
        Return: Token() object
            The identified  number token.
        '''
        return Tokenizer.get_token_on_first_match(input=input, type=TokenType.NUMBER, regex=self.NUMBER_REGEX)

    def get_operator_token(self, input):
        '''
        Find token of type TokenType.OPERATOR in the query.

        Parameters
        input: string
            The query to format.
        
        Return: Token() object
            The identified operator token.
        '''
        return Tokenizer.get_token_on_first_match(input=input, type=TokenType.OPERATOR, regex=self.OPERATOR_REGEX)

    def get_keyword_token(self, input, previousToken):
        """
        A keyword cannot be preceded by a "."
        This makes it so in "my_table.from", "from" is not considered a key word

        Find keyword token in the query.

        Parameters
        input: string
            The query to format.
        previousToken: Token() object
            The last identified token.
        
        Return: Token() object
            The identified keyword token.
        """
        if (previousToken and previousToken.value and previousToken.value == '.'):
            return
        return (self.get_top_level_keyword_token(input) or self.get_newline_keyword_token(input)
                or self.get_top_level_keyword_token_no_indent(input) or self.get_reserved_keyword_token(input)
                or self.get_plain_keyword_token(input))

    def get_reserved_keyword_token(self, input):
        '''
        Find token of type TokenType.RESERVED_KEYWORD in the query.

        Parameters
        input: string
            The query to format.
        
        Return: Token() object
            The identified reserved keyword token.
        '''
        return Tokenizer.get_token_on_first_match(input=input,
                                                  type=TokenType.RESERVED_KEYWORD,
                                                  regex=self.RESERVED_KEYWORD_REGEX)

    def get_top_level_keyword_token(self, input):
        '''
        Find token of type TokenType.TOP_LEVEL_KEYWORD in the query.

        Parameters
        input: string
            The query to format.
        
        Return: Token() object
            The identified top-level keyword token.
        '''
        return Tokenizer.get_token_on_first_match(input=input,
                                                  type=TokenType.TOP_LEVEL_KEYWORD,
                                                  regex=self.TOP_LEVEL_KEYWORD_REGEX)

    def get_newline_keyword_token(self, input):
        '''
        Find token of type TokenType.NEWLINE_KEYWORD in the query.

        Parameters
        input: string
            The query to format.
        
        Return: Token() object
            The identified newline keyword token.
        '''
        return Tokenizer.get_token_on_first_match(input=input,
                                                  type=TokenType.NEWLINE_KEYWORD,
                                                  regex=self.NEWLINE_KEYWORD_REGEX)

    def get_top_level_keyword_token_no_indent(self, input):
        '''
        Find token of type TokenType.TOP_LEVEL_KEYWORD_TOKEN_NO_INDENT in the query.

        Parameters
        input: string
            The query to format.
        
        Return: Token() object
            The identified token.
        '''
        return Tokenizer.get_token_on_first_match(input=input,
                                                  type=TokenType.TOP_LEVEL_KEYWORD_NO_INDENT,
                                                  regex=self.TOP_LEVEL_KEYWORD_NO_INDENT_REGEX)

    def get_plain_keyword_token(self, input):
        '''
        Find token of type TokenType.KEYWORD in the query.

        Parameters
        input: string
            The query to format.
        
        Return: Token() object
            The identified keyword token.
        '''
        return Tokenizer.get_token_on_first_match(input=input, type=TokenType.KEYWORD, regex=self.PLAIN_KEYWORD_REGEX)

    def get_word_token(self, input):
        '''
        Find token of type TokenType.WORD in the query.

        Parameters
        input: string
            The query to format.
        
        Return: Token() object
            The identified word token.
        '''
        return Tokenizer.get_token_on_first_match(input=input, type=TokenType.WORD, regex=self.WORD_REGEX)

    @staticmethod
    def get_token_on_first_match(input, type, regex):
        '''
        Find token of given type upon the first match.

        Parameters
        input: string
            The query to format.
        type: string
            Type of the token.
        regex: string
            The regex pattern that matches the token.
        
        Return: Token() object
            The matched token.
        '''
        if type in [
                TokenType.RESERVED_KEYWORD, TokenType.TOP_LEVEL_KEYWORD, TokenType.TOP_LEVEL_KEYWORD_NO_INDENT,
                TokenType.NEWLINE_KEYWORD, TokenType.KEYWORD, TokenType.OPEN_PAREN, TokenType.CLOSE_PAREN
        ]:
            # matches = re.search(pattern=regex, string=input, flags=re.IGNORECASE | re.UNICODE)
            matches = re.match(pattern=regex, string=input, flags=re.IGNORECASE | re.UNICODE)
        else:
            # matches = re.search(pattern=regex, string=input, flags=re.UNICODE)
            matches = re.match(pattern=regex, string=input, flags=re.UNICODE)
        if matches:
            return Token(type=type, value=matches.group(0))



# -*- coding: utf-8 -*-
# MIT License

# Copyright (c) 2016-present ZeroTurnaround LLC
# Copyright (c) 2016-present kufii
# Copyright (c) 2020-present largecats

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from sparksqlformatter.src.tokenizer import TokenType


class InlineBlock:
    '''
    Class for managing inline blocks.

    Inline blocks are parenthesized expressions that are shorter than self.inlineMaxLength.
    These blocks are formatted on a single line, unlike longer parenthesized expressions where open-parenthesis causes newline and increase of indentation.
    '''
    def __init__(self, inlineMaxLength):
        self.level = 0
        self.inlineMaxLength = inlineMaxLength

    def begin_if_possible(self, tokens, index):
        """
        Begins inline block when lookahead through upcoming tokens determines that the block would be smaller than self.inlineMaxLength.

        Parameters
        tokens: list
            List of tokens.
        index: int
            Current token position.
        """
        if self.level == 0 and self.is_inline_block(tokens, index):
            self.level = 1
        elif self.level > 0:
            self.level += 1
        else:
            self.level = 0

    def end(self):
        """
        Finishes current inline block. There might be several nested ones.
        """
        self.level -= 1

    def is_active(self):
        """
        Check if currently is inside an inline block.

        Return: bool
            Returns True when inside an inline block.
        """
        return self.level > 0

    def is_inline_block(self, tokens, index):
        """
        Check if this should be an inline parentheses block.
        Examples are "NOW()", "COUNT(*)", "int(10)", key(`some_column`), DECIMAL(7,2).

        Parameters
        tokens: list
            List of tokens.
        index: int
            Current token position.
        """
        length = 0
        level = 0

        for i in range(index, len(tokens)):
            token = tokens[i]
            length += len(token.value)

            if length > self.inlineMaxLength:
                return False

            if token.type == TokenType.OPEN_PAREN:
                level += 1
            elif token.type == TokenType.CLOSE_PAREN:
                level -= 1
                if level == 0:
                    return True

            if InlineBlock.is_forbidden_token(token):
                return False

        return False

    @staticmethod
    def is_forbidden_token(token):
        '''
        Check if token does not belong to an inline block.

        Parameters
        token: sparksqlformatter.src.tokenizer.Token() object

        Return: bool
        '''
        type, value = token.type, token.value
        return (type == TokenType.TOP_LEVEL_KEYWORD or type == TokenType.NEWLINE_KEYWORD
                or type == TokenType.LINE_COMMENT or type == TokenType.BLOCK_COMMENT or value == ';')



# -*- coding: utf-8 -*-
# MIT License

# Copyright (c) 2016-present ZeroTurnaround LLC
# Copyright (c) 2016-present kufii
# Copyright (c) 2020-present largecats

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
INDENT_TYPE_TOP_LEVEL = 'top-level'  # increased by open parentheses
INDENT_TYPE_BLOCK_LEVEL = 'block-level'  # increased by TOPLEVEL_KEYWORD tokens


class Indentation:
    '''
    Class for managing indentation.
    '''
    def __init__(self, indent):
        '''
        Parameters
        indent: string
            One unit of indentation.
        '''
        self.indent = indent or '   '
        self.indentTypes = []  # store each indentation that needs to be applied

    def get_indent(self):
        """
        Get indentation at current level.

        Return: string
            The current level indentation.
        """
        return self.indent * len(self.indentTypes)

    def increase_top_level(self):
        """
        Increases indentation by one top-level indent.
        """
        self.indentTypes.append(INDENT_TYPE_TOP_LEVEL)

    def increase_block_level(self):
        """
        Increases indentation by one block-level indent.
        """
        self.indentTypes.append(INDENT_TYPE_BLOCK_LEVEL)

    def decrease_top_level(self):
        """
        Decreases indentation by one top-level indent.
        Does nothing if the previous indent is not top-level.
        """
        if len(self.indentTypes) > 0:
            if (self.indentTypes[-1] == INDENT_TYPE_TOP_LEVEL):
                self.indentTypes.pop()

    def decrease_block_level(self):
        """
        Decreases indentation by one block-level indent.
        If there are top-level indents within the block-level indent, throws away these as well.
        """
        while len(self.indentTypes) > 0:
            type = self.indentTypes.pop()
            if (type != INDENT_TYPE_TOP_LEVEL):
                break

    def reset_indentation(self):
        '''
        Reset indentation level by clearing self.indentTypes.
        '''
        self.indentTypes = []



# -*- coding: utf-8 -*-
# MIT License

# Copyright (c) 2016-present ZeroTurnaround LLC
# Copyright (c) 2016-present kufii
# Copyright (c) 2020-present largecats

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from sparksqlformatter.src import config

DEFAULT_STYLE_SECTION = 'sparksqlformatter'  # default section heading for style config files


class Style:
    '''
    Class for formatting style.
    '''
    def __init__(
            self,
            topLevelKeywords=config.Keyword.TOP_LEVEL_KEYWORDS,
            topLevelKeywordsNoIndent=config.Keyword.TOP_LEVEL_KEYWORDS_NO_INDENT,
            newlineKeywords=config.Keyword.NEWLINE_KEYWORDS,
            userDefinedFunctions=[],
            stringTypes=['""', "''", '{}'],
            openParens=['(', '[', 'CASE'],
            closeParens=[')', ']', 'END'],  # the order of the parentheses need to match with openParens
            lineCommentTypes=['--'],
            reservedKeywordUppercase=True,
            linesBetweenQueries=1,
            specialWordChars=[],
            indent='    ',
            inlineMaxLength=120,
            splitOnComma=True):
        '''
        Parameters
        topLevelKeywords: list
            Keywords that initiate top-level blocks in the query so that the following lines are indented.
            E.g., SELECT and FROM in
            SELECT
                ...
            FROM
                ...
            Default to config.Keyword.TOP_LEVEL_KEYWORDS.
        topLevelKeywordsNoIndent: list
            Keywords that initiate top-level blocks in the query without indenting the following lines.
            E.g., UNION in
            SELECT
                *
            FROM
                ...
            UNION
            SELECT
                *
            FROM
                ...
            Default to config.Keyword.TOP_LEVEL_KEYWORDS_NO_INDENT.
        newlineKeywords: list
            Keywords that initiate a newline in the query.
            Default to config.Keyword.NEWLINE_KEYWORDS.
        userDefinedFunctions: list
            Functions defined by user. To be treated as in-built functions in formatting.
            Default to [].
        stringTypes: list
            Pairs of characters that enclose strings in the query.
            Default to ['""', "''", '{}'].
        openParens: list
            Characters tbat behave as opening parentheses in the query.
            Default to ['(', 'CASE']
        closdParens: list
            Characters that behave as closing parentheses in the query. 
            The order of the closing parentheses need to match with that of the opening parentheses.
            Default to [')', 'END']
        lineCommentTypes: list
            Strings that initiate comments in the query.
            Default to ['--']
        resercedKeywordsUppercase: bool
            If True, convert all reserved keywords to uppercase.
            Else, convert all reserved keywords to lowercase.
            Default to True.
        linesBetweenQueries: int
            Number of blank lines to put between (sub)queries.
            Default to 1.
        specialWordChars: list
            Characters with special meanings in the query lanauge.
            Default to [].
        indent: string
            One unit of indentation.
            Default to '    '.
        inlineMaxLength: int
            Maximum length of an inline block.
            Default to 120.
        splitOnComma: bool
           If a comma separated list (e.g., in GROUP BY, ORDER BY) is too long to fit in a line, split such that all elements are on a single line.
           Else, will only split at inlineMaxLength.
        # '''
        # self.keywords = (config.Keyword.RESERVED_KEYWORDS + config.Keyword.NON_RESERVED_KEYWORDS +
        #                  config.Function.AGGREGATE_FUNCTIONS + config.Function.ARRAY_FUNCTIONS +
        #                  config.Function.CONDITIONAL_FUNCTIONS + config.Function.DATE_TIME_FUNCTIONS +
        #                  config.Function.HASH_FUNCTIONS + config.Function.JSON_FUNCTIONS + config.Function.MAP_FUNCTIONS +
        #                  config.Function.MATHEMATICAL_FUNCTIONS + config.Function.MISC_FUNCTIONS +
        #                  config.Function.OPERATOR_FUNCTIONS + config.Function.STRING_FUNCTIONS + config.Function.STRUCT_FUNCTIONS +
        #                  config.Function.TABLE_GENERATING_FUNCTIONS + config.Function.TYPE_CONVERSION_FUNCTIONS +
        #                  config.Function.WINDOW_FUNCTIONS + config.Function.XPATH_FUNCTIONS + userDefinedFunctions)
        # self.reservedKeywords = config.Keyword.RESERVED_KEYWORDS
        self.topLevelKeywords = topLevelKeywords
        self.newlineKeywords = newlineKeywords
        self.topLevelKeywordsNoIndent = topLevelKeywordsNoIndent
        self.userDefinedFunctions = userDefinedFunctions
        self.stringTypes = stringTypes
        self.openParens = openParens
        self.closeParens = closeParens
        self.lineCommentTypes = lineCommentTypes
        self.reservedKeywordUppercase = reservedKeywordUppercase
        self.linesBetweenQueries = linesBetweenQueries
        self.specialWordChars = specialWordChars
        self.indent = indent
        self.inlineMaxLength = inlineMaxLength
        self.splitOnComma = splitOnComma



# -*- coding: utf-8 -*-
# MIT License

# Copyright (c) 2020-present largecats

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# from __future__ import print_function  # for print() in Python 2
from io import open  # for open() in Python 2
import sys
import re
import logging
if sys.version_info[0] >= 3:
    import configparser
else:
    from backports import configparser
import ast

from sparksqlformatter.src.style import Style
from sparksqlformatter.src.formatter import Formatter
from sparksqlformatter.src.style import DEFAULT_STYLE_SECTION

logger = logging.getLogger(__name__)
log_formatter = '[%(asctime)s] %(levelname)s [%(filename)s:%(lineno)s:%(funcName)s] %(message)s'
logging.basicConfig(stream=sys.stdout, level=logging.INFO, format=log_formatter)


def format_file(filePath, style=Style(), inPlace=False):
    '''
    Format given file with given styleurations.

    Parameters
    filePath: string
        Path to the file to format.
    style: string, dict, or sparksqlformatter.src.style.Style() object
        Styleurations for the query language.
    inPlace: bool
        If True, will format the file in place.
        Else, will write the formatted file to stdout.
    '''
    if type(style) == type(Style()):  # style is a Style() object
        formatter = Formatter(style=style)
    else:  # create Style() object from style
        if type(style) == str:
            if style.startswith('{'):  # style is a dictionary in string
                style = eval(style)
                formatter = Formatter(style=_create_style_from_dict(style))
            else:  # style is a file path
                formatter = Formatter(style=_create_style_from_file(style))
        elif type(style) == dict:  # style is a dictionary
            formatter = Formatter(style=_create_style_from_dict(style))
        else:
            raise Exception('Unsupported style type')
    _format_file(filePath, formatter, inPlace)


def format_query(query, style=Style()):
    '''
    Format query using given styleurations.

    Parameters
    query: string
        The query to be formatted.
    style: string, dict, or sparksqlformatter.src.style.Style() object
        Styleurations for the query language.
    
    Return: string
        The formatted query.
    '''
    if type(style) == type(Style()):
        formatter = Formatter(style=style)
    else:
        if type(style) == str:
            if style.startswith('{'):
                style = eval(style)
                formatter = Formatter(style=_create_style_from_dict(style))
            else:
                formatter = Formatter(style=_create_style_from_file(style))
        elif type(style) == dict:
            formatter = Formatter(style=_create_style_from_dict(style))
        else:
            raise Exception('Unsupported style type')
    return _format_query(query, formatter)


def _format_file(filePath, formatter, inPlace=False):
    '''
    The I/O helper function for format_file(). Read from given file, format it, and write to specified output.

    Parameters
    filePath: string
        Path to the file to format.
    formatter: sparksqlformatter.src.formatter.Formatter() object
        Formatter.
    inPlace: bool
        If True, will format the file in place.
        Else, will write the formatted file to stdout.
    '''
    query = _read_from_file(filePath)
    formattedQuery = _format_query(query, formatter)
    if inPlace:  # overwrite file
        logger.info('Writing to ' + filePath + '...')
        _write_to_file(formattedQuery, filePath)
    else:  # write to stdout
        sys.stdout.write(formattedQuery)


def _read_from_file(filePath):
    '''
    The input helper function for _format_file(). Read from given file and return its content.

    Parameters
    filePath: string
        Path to the file to format.
    
    Return: string
        The file content.
    '''
    # see https://docs.python.org/3.5/library/functions.html#open
    with open(file=filePath, mode='r', newline=None, encoding='utf-8') as f:
        text = f.read()
    return text


def _write_to_file(formattedQuery, filePath):
    '''
    The output helper function for _format_file(). Write formatted query to given file.

    Parameters
    formattedQuery: string
        The formatted query.
    filePath: string
        Path to the file to write to.
    '''
    # see https://docs.python.org/3.5/library/functions.html#open
    with open(file=filePath, mode='w', newline='\n', encoding='utf-8') as f:
        f.write(formattedQuery)


def _format_query(query, formatter):
    '''
    The wrapper function for format_query(). Format a given query using given formatter.

    Parameters
    query: string
        The query to format.
    formatter: sparksqlformatter.src.formatter.Formatter object
        Formatter.
    
    Return: string
        The formatted query
    '''
    return formatter.format(query)


def _create_style_from_dict(styleDict, defaultStyleSection=DEFAULT_STYLE_SECTION):
    '''
    Create Style() object from dictionary.

    Parameters
    styleDict: dict
        A dictionary of styleurations specified in key-value pairs.
    defaultStyleSection: string
        The top-level style section that needs to be added on top of the styleDict before feeding to styleParser.read_dict(), default to 'sparksqlformatter'.
    
    Return: sparksqlformatter.src.style.Style() object
        The Style() object created from styleDict.
    '''
    styleParser = configparser.ConfigParser()
    styleParser.optionxform = str  # makes the parser case-sensitive
    if defaultStyleSection not in styleDict:
        styleDict = {defaultStyleSection: styleDict}  # add top-level section
    styleParser.read_dict(styleDict)  # styleParser assumes the existence of a top-level section
    args = _parse_args_in_correct_type(styleParser, defaultStyleSection)
    style = Style(**args)
    return style


def _create_style_from_file(styleFilePath, defaultStyleSection=DEFAULT_STYLE_SECTION):
    '''
    Create Style() object from style file.

    Parameters
    styleFilePath: string
        Path to the style file.
    defaultStyleSection: string
        The top-level style section that needs to be added on top of the styleDict before feeding to styleParser.read_dict(), default to 'sparksqlformatter'.
    
    Return: sparksqlformatter.src.style.Style() object
        The Style() object created from the style file.
    '''
    styleParser = configparser.ConfigParser()
    styleParser.optionxform = str  # makes the parser case-sensitive
    styleParser.read(styleFilePath)
    if defaultStyleSection in styleParser:
        styleDict = _parse_args_in_correct_type(styleParser, defaultStyleSection)
        return Style(**styleDict)
    else:
        raise Exception('Section ' + defaultStyleSection + 'not found in ' + styleFilePath)


def _parse_args_in_correct_type(styleParser, defaultStyleSection=DEFAULT_STYLE_SECTION):
    '''
    Parse paramters in style with special handling to convert boolean values converted from string back to boolean type.

    Parameters
    styleParser: a styleParser.ConfigParser() object
        Parser for style files.
    defaultStyleSection: string
        The top-level style section that needs to be added on top of the styleDict before feeding to styleParser.read_dict(), default to 'sparksqlformatter'.
    
    Return: dict
        A dictionary of styleuration key-value pairs where values are of correct type.
    '''
    args = {}
    for key in styleParser[defaultStyleSection]:
        args[key] = ast.literal_eval(styleParser[defaultStyleSection][key])
    return args



# -*- coding: utf-8 -*-


class SparkSQL:
    # http://spark.apache.org/docs/latest/sql-ref-ansi-compliance.html#sql-keywords
    class Keyword:
        RESERVED_KEYWORDS = [
            'ALL', 'ANALYZE', 'AND', 'ANY', 'ARRAY', 'AS', 'ASC', 'AT', 'BETWEEN', 'BOTH', 'BY', 'CASE', 'CAST',
            'CHECK', 'COLLATE', 'COLUMN', 'CONSTRAINT', 'CREATE', 'CROSS', 'CURRENT', 'CURRENT_DATE', 'CURRENT_TIME',
            'CURRENT_TIMESTAMP', 'CURRENT_USER', 'DATABASE', 'DAY', 'DELETE', 'DESC', 'DESCRIBE', 'DISTINCT', 'DIV',
            'DROP', 'ELSE', 'END', 'ESCAPE', 'EXCEPT', 'EXISTS', 'EXTERNAL', 'EXTRACT', 'FALSE', 'FETCH', 'FILTER',
            'FOLLOWING', 'FOR', 'FOREIGN', 'FROM', 'FULL', 'FUNCTION', 'GLOBAL', 'GRANT', 'GROUP', 'GROUPING', 'HAVING',
            'HOUR', 'IF', 'IN', 'INNER', 'INSERT', 'INTERSECT', 'INTERVAL', 'INTO', 'IS', 'JOIN', 'KEYS', 'LATERAL',
            'LEADING', 'LEFT', 'LIKE', 'LIMIT', 'LOCAL', 'MINUS', 'MINUTE', 'MONTH', 'NATURAL', 'NO', 'NOT', 'NULL',
            'OF', 'ON', 'ONLY', 'OPTION', 'OPTIONS', 'OR', 'ORDER', 'OUT', 'OUTER', 'OVER', 'OVERLAPS', 'PARTITION',
            'POSITION', 'PRIMARY', 'REFERENCES', 'REPAIR', 'RIGHT', 'ROW', 'ROWS', 'SCHEMA', 'SECOND', 'SELECT', 'SEMI',
            'SESSION_USER', 'SET', 'SKEWED', 'SOME', 'SORT', 'START', 'TABLE', 'THEN', 'TO', 'TRAILING', 'TRUE',
            'TRUNCATE', 'UNBOUNDED', 'UNION', 'UNIQUE', 'UNKNOWN', 'UPDATE', 'USE', 'USER', 'USING', 'VALUES', 'VIEW',
            'WHEN', 'WHERE', 'WINDOW', 'WITH', 'YEAR'
        ]

        NON_RESERVED_KEYWORDS = [
            'ADD', 'AFTER', 'ALTER', 'ANTI', 'ARCHIVE', 'AUTHORIZATION', 'BUCKET', 'BUCKETS', 'CACHE', 'CASCADE',
            'CHANGE', 'CLEAR', 'CLUSTER', 'CLUSTERED', 'CODEGEN', 'COLLECTION', 'COLUMNS', 'COMMENT', 'COMMIT',
            'COMPACT', 'COMPACTIONS', 'COMPUTE', 'CONCATENATE', 'COST', 'CUBE', 'DATA', 'DATABASES', 'DBPROPERTIES',
            'DEFINED', 'DELIMITED', 'DFS', 'DIRECTORIES', 'DIRECTORY', 'DISTRIBUTE', 'ESCAPED', 'EXCHANGE', 'EXPLAIN',
            'EXPORT', 'EXTENDED', 'FIELDS', 'FILEFORMAT', 'FIRST', 'FORMAT', 'FORMATTED', 'FUNCTIONS', 'IGNORE',
            'IMPORT', 'INDEX', 'INDEXES', 'INPATH', 'INPUTFORMAT', 'ITEMS', 'LAST', 'LAZY', 'LINES', 'LIST', 'LOAD',
            'LOCATION', 'LOCK', 'LOCKS', 'LOGICAL', 'MACRO', 'MAP', 'MATCHED', 'MERGE', 'MSCK', 'NAMESPACE',
            'NAMESPACES', 'NULLS', 'OUTPUTFORMAT', 'OVERLAY', 'OVERWRITE', 'PARTITIONED', 'PARTITIONS', 'PERCENT',
            'PIVOT', 'PLACING', 'PRECEDING', 'PRINCIPALS', 'PROPERTIES', 'PURGE', 'QUERY', 'RECORDREADER',
            'RECORDWRITER', 'RECOVER', 'REDUCE', 'REFRESH', 'RENAME', 'REPLACE', 'RESET', 'RESTRICT', 'REVOKE', 'RLIKE',
            'ROLE', 'ROLES', 'ROLLBACK', 'ROLLUP', 'SEPARATED', 'SERDE', 'SERDEPROPERTIES', 'SETS', 'SHOW', 'SORTED',
            'STATISTICS', 'STORED', 'STRATIFY', 'STRUCT', 'SUBSTR', 'SUBSTRING', 'TABLES', 'TABLESAMPLE',
            'TBLPROPERTIES', 'TEMPORARY', 'TERMINATED', 'TOUCH', 'TRANSACTION', 'TRANSACTIONS', 'TRANSFORM', 'TRIM',
            'UNARCHIVE', 'UNCACHE', 'UNLOCK', 'UNSET', 'VIEWS'
        ]

        DATA_TYPES = [
            'BOOLEAN', 'BYTE', 'TINYINT', 'SHORT', 'SMALLINT', 'INT', 'INTEGER', 'LONG', 'BIGINT', 'FLOAT', 'REAL',
            'DOUBLE', 'DATE', 'TIMESTAMP', 'STRING', 'BINARY', 'DECIMAL', 'DEC', 'NUMERIC', 'INTERVAL', 'ARRAY',
            'STRUCT', 'MAP'
        ]

        TOP_LEVEL_KEYWORDS = [
            'ADD', 'AFTER', 'ALTER COLUMN', 'ALTER TABLE', 'CREATE TABLE', 'CROSS JOIN', 'DELETE FROM', 'EXCEPT',
            'FETCH FIRST', 'FROM', 'GROUP BY', 'GO', 'HAVING', 'INNER JOIN', 'INSERT INTO', 'INSERT', 'JOIN',
            'LEFT JOIN', 'LEFT OUTER JOIN', 'LIMIT', 'MODIFY', 'ORDER BY', 'OUTER JOIN', 'PARTITION BY', 'RIGHT JOIN',
            'RIGHT OUTER JOIN', 'SELECT', 'SET CURRENT SCHEMA', 'SET SCHEMA', 'SET', 'UPDATE', 'VALUES', 'WHERE'
        ]

        TOP_LEVEL_KEYWORDS_NO_INDENT = ['INTERSECT', 'INTERSECT ALL', 'MINUS', 'UNION', 'UNION ALL']

        NEWLINE_KEYWORDS = [
            'AND', 'ELSE', 'LATERAL', 'ON', 'OPTIONS', 'OR', 'PARTITIONED BY', 'THEN', 'USING', 'WHEN', 'XOR'
        ]

    class Function:
        # http://spark.apache.org/docs/latest/sql-ref-functions-builtin.html
        AGGREGATE_FUNCTIONS = [
            'any', 'approx_count_distinct', 'approx_percentile', 'avg', 'bit_or', 'bit_xor', 'bool_and', 'bool_or',
            'collect_list', 'collect_set', 'corr', 'count', 'count_if', 'count_min_sketch', 'covar_pop', 'covar_samp',
            'every', 'first', 'first_value', 'kurtosis', 'last', 'last_value', 'max', 'mean', 'min', 'min_by',
            'percentile', 'skewness', 'some', 'std', 'stddev', 'stddev_pop', 'stddev_samp', 'sum', 'var_pop',
            'var_samp', 'variance'
        ]
        ARRAY_FUNCTIONS = [
            'array_contains', 'array_distinct', 'array_except', 'array_intersect', 'array_join', 'array_max',
            'array_min', 'array_position', 'array_remove', 'array_repeat', 'array_union', 'arrays_overlap',
            'arrays_zip', 'concat', 'flatten', 'reverse', 'sequence', 'shuffle', 'slice', 'sort_array'
        ]
        DATE_TIME_FUNCTIONS = [
            'add_months', 'current_date', 'current_timestamp', 'date_add', 'date_format', 'date_part', 'date_sub',
            'date_trunc', 'datediff', 'dayofweek', 'dayofyear', 'from_unixtime', 'from_utc_timestamp', 'hour',
            'last_day', 'make_date', 'make_timestamp', 'minute', 'month', 'months_between', 'next_day', 'now',
            'quarter', 'second', 'to_date', 'to_timestamp', 'to_unix_timestamp', 'to_utc_timestamp', 'trunc',
            'unix_timestamp', 'weekday', 'weekofyear', 'year'
        ]
        JSON_FUNCTIONS = ['from_json', 'get_json_object', 'json_tuple', 'schema_of_json', 'to_json']
        MAP_FUNCTIONS = ['map_concat', 'map_entries', 'map_from_entries', 'map_keys', 'map_values']
        WINDOW_FUNCTIONS = ['cume_dist', 'dense_rank', 'lag', 'lead', 'ntile', 'percent_rank', 'rank', 'row_number']


class SQL:
    # http://spark.apache.org/docs/latest/api/sql/index.html
    class Function:
        AGGREGATE_FUNCTIONS = [
            'aggregate', 'any', 'approx_count_distinct', 'approx_percentile', 'avg', 'bit_and', 'bit_count',
            'bit_length', 'bit_or', 'bit_xor', 'bool_and', 'bool_or', 'collect_list', 'collect_set', 'corr', 'count',
            'count_if', 'count_min_sketch', 'covar_pop', 'covar_samp', 'every', 'first', 'first_value', 'greatest',
            'kurtosis', 'last', 'last_value', 'least', 'length', 'max', 'max_by', 'mean', 'min', 'min_by',
            'octet_length', 'percentile', 'percentile_approx', 'size', 'skewness', 'some', 'std', 'stddev',
            'stddev_pop', 'stddev_samp', 'sum', 'var_pop', 'var_samp', 'variance'
        ]
        ARRAY_FUNCTIONS = [
            'array', 'array_contains', 'array_distinct', 'array_except', 'array_intersect', 'array_join', 'array_max',
            'array_min', 'array_position', 'array_remove', 'array_repeat', 'array_sort', 'array_union',
            'arrays_overlap', 'arrays_zip', 'cardinality', 'element_at', 'elt', 'exists', 'filter', 'find_in_set',
            'flatten', 'forall', 'reverse', 'sequence', 'shuffle', 'slice', 'sort_array', 'transform', 'zip_with'
        ]
        CONDITIONAL_FUNCTIONS = [
            'assert_true', 'coalesce', 'if', 'ifnull', 'in', 'isnan', 'isnotnull', 'isnull', 'nanvl', 'nullif', 'nvl',
            'nvl2', 'when'
        ]
        DATE_TIME_FUNCTIONS = [
            'add_months', 'current_date', 'current_timestamp', 'date', 'date_add', 'date_format', 'date_part',
            'date_sub', 'date_trunc', 'datediff', 'day', 'dayofmonth', 'dayofweek', 'dayofyear', 'extract',
            'from_unixtime', 'from_utc_timestamp', 'hour', 'last_day', 'make_date', 'make_interval', 'make_timestamp',
            'minute', 'month', 'months_between', 'next_day', 'now', 'quarter', 'second', 'timestamp', 'to_date',
            'to_timestamp', 'to_unix_timestamp', 'to_utc_timestamp', 'trunc', 'unix_timestamp', 'weekday', 'weekofyear',
            'year'
        ]
        HASH_FUNCTIONS = ['hash', 'crc32', 'sha', 'sha1', 'sha2']
        JSON_FUNCTIONS = ['get_json_object', 'schema_of_json', 'json_tuple', 'from_json', 'to_json']
        MAP_FUNCTIONS = [
            'map', 'map_concat', 'map_entries', 'map_filter', 'map_from_arrays', 'map_from_entries', 'map_keys',
            'map_values', 'map_zip_with', 'str_to_map', 'transform_keys', 'transform_values'
        ]
        MATHEMATICAL_FUNCTIONS = [
            'abs', 'acos', 'acosh', 'asin', 'asinh', 'atan', 'atan2', 'atanh', 'bround', 'cbrt', 'ceil', 'ceiling',
            'conv', 'cos', 'cosh', 'cot', 'degrees', 'div', 'e', 'exp', 'expm1', 'factorial', 'floor', 'format_number',
            'hypot', 'ln', 'log', 'log10', 'log1p', 'log2', 'mod', 'pi', 'pmod', 'pow', 'power', 'radians', 'rint',
            'round', 'sign', 'signum', 'sin', 'sinh', 'sqrt', 'tan', 'tanh'
        ]
        MISC_FUNCTIONS = [
            'current_database', 'input_file_block_length', 'input_file_block_start', 'input_file_name', 'grouping',
            'grouping_id', 'java_method', 'monotonically_increasing_id', 'positive', 'rand', 'randn', 'random',
            'reflect', 'cube', 'rollup', 'schema_of_csv', 'spark_partition_id', 'typeof', 'uuid', 'version', 'xxhash64'
        ]
        OPERATOR_FUNCTIONS = ['negative', 'not', 'or', 'shiftleft', 'shiftright', 'shiftrightunsigned']
        STRING_FUNCTIONS = [
            'ascii', 'base64', 'char', 'char_length', 'character_length', 'chr', 'concat', 'concat_ws', 'decode',
            'encode', 'format_string', 'initcap', 'instr', 'lcase', 'left', 'levenshtein', 'like', 'locate', 'lower',
            'lpad', 'ltrim', 'md5', 'overlay', 'parse_url', 'position', 'printf', 'regexp_extract', 'regexp_replace',
            'repeat', 'replace', 'right', 'rlike', 'rpad', 'rtrim', 'sentences', 'soundex', 'space', 'split', 'substr',
            'substring', 'substring_index', 'translate', 'trim', 'ucase', 'unbase64', 'upper'
        ]
        STRUCT_FUNCTIONS = ['from_csv', 'named_struct', 'struct', 'to_csv']
        TABLE_GENERATING_FUNCTIONS = [
            'explode', 'explode_outer', 'inline', 'stack', 'inline_outer', 'posexplode', 'posexplode_outer'
        ]
        TYPE_CONVERSION_FUNCTIONS = [
            'bigint', 'bin', 'binary', 'boolean', 'cast', 'decimal', 'double', 'float', 'hex', 'int', 'smallint',
            'string', 'tinyint', 'unhex'
        ]
        WINDOW_FUNCTIONS = ['cume_dist', 'dense_rank', 'lag', 'lead', 'ntile', 'percent_rank', 'rank', 'row_number']
        XPATH_FUNCTIONS = [
            'xpath', 'xpath_boolean', 'xpath_double', 'xpath_float', 'xpath_int', 'xpath_long', 'xpath_number',
            'xpath_short', 'xpath_string'
        ]


class Keyword:
    RESERVED_KEYWORDS = SparkSQL.Keyword.RESERVED_KEYWORDS
    NON_RESERVED_KEYWORDS = SparkSQL.Keyword.NON_RESERVED_KEYWORDS + SparkSQL.Keyword.DATA_TYPES
    TOP_LEVEL_KEYWORDS = SparkSQL.Keyword.TOP_LEVEL_KEYWORDS
    TOP_LEVEL_KEYWORDS_NO_INDENT = SparkSQL.Keyword.TOP_LEVEL_KEYWORDS_NO_INDENT
    NEWLINE_KEYWORDS = SparkSQL.Keyword.NEWLINE_KEYWORDS


class Function:
    AGGREGATE_FUNCTIONS = list(set(SparkSQL.Function.AGGREGATE_FUNCTIONS + SQL.Function.AGGREGATE_FUNCTIONS))
    ARRAY_FUNCTIONS = list(set(SparkSQL.Function.ARRAY_FUNCTIONS + SQL.Function.ARRAY_FUNCTIONS))
    CONDITIONAL_FUNCTIONS = SQL.Function.CONDITIONAL_FUNCTIONS
    DATE_TIME_FUNCTIONS = list(set(SparkSQL.Function.DATE_TIME_FUNCTIONS + SQL.Function.DATE_TIME_FUNCTIONS))
    HASH_FUNCTIONS = SQL.Function.HASH_FUNCTIONS
    JSON_FUNCTIONS = list(set(SparkSQL.Function.JSON_FUNCTIONS + SQL.Function.JSON_FUNCTIONS))
    MAP_FUNCTIONS = list(set(SparkSQL.Function.MAP_FUNCTIONS + SQL.Function.MAP_FUNCTIONS))
    MATHEMATICAL_FUNCTIONS = SQL.Function.MATHEMATICAL_FUNCTIONS
    MISC_FUNCTIONS = SQL.Function.MISC_FUNCTIONS
    OPERATOR_FUNCTIONS = SQL.Function.OPERATOR_FUNCTIONS
    STRING_FUNCTIONS = list(set(SparkSQL.Function.JSON_FUNCTIONS + SQL.Function.STRING_FUNCTIONS))
    STRUCT_FUNCTIONS = SQL.Function.STRUCT_FUNCTIONS
    TABLE_GENERATING_FUNCTIONS = SQL.Function.TABLE_GENERATING_FUNCTIONS
    TYPE_CONVERSION_FUNCTIONS = SQL.Function.TYPE_CONVERSION_FUNCTIONS
    WINDOW_FUNCTIONS = list(set(SparkSQL.Function.WINDOW_FUNCTIONS + SQL.Function.WINDOW_FUNCTIONS))
    XPATH_FUNCTIONS = SQL.Function.XPATH_FUNCTIONS


KEYWORDS = (Keyword.RESERVED_KEYWORDS + Keyword.NON_RESERVED_KEYWORDS + Function.AGGREGATE_FUNCTIONS +
            Function.ARRAY_FUNCTIONS + Function.CONDITIONAL_FUNCTIONS + Function.DATE_TIME_FUNCTIONS +
            Function.HASH_FUNCTIONS + Function.JSON_FUNCTIONS + Function.MAP_FUNCTIONS +
            Function.MATHEMATICAL_FUNCTIONS + Function.MISC_FUNCTIONS + Function.OPERATOR_FUNCTIONS +
            Function.STRING_FUNCTIONS + Function.STRUCT_FUNCTIONS + Function.TABLE_GENERATING_FUNCTIONS +
            Function.TYPE_CONVERSION_FUNCTIONS + Function.WINDOW_FUNCTIONS + Function.XPATH_FUNCTIONS)



# -*- coding: utf-8 -*-
# MIT License

# Copyright (c) 2020-present largecats

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# from __future__ import print_function  # for print() in Python 2
import os
import sys
import argparse
import logging
import codecs
import json
sys.path.insert(0, os.path.dirname(os.path.dirname(os.path.abspath(__file__))))

from sparksqlformatter.src.style import Style
from sparksqlformatter.src import api
from sparksqlformatter.src.formatter import Formatter

logger = logging.getLogger(__name__)
log_formatter = '[%(asctime)s] %(levelname)s [%(filename)s:%(lineno)s:%(funcName)s] %(message)s'
logging.basicConfig(stream=sys.stdout, level=logging.INFO, format=log_formatter)


def main(argv):
    '''
    Main function that enables formatting file from command-line.

    Parameters
    argv: list
        List of arguments in sys.argv, excluding the first argument which is the script itself.
    '''
    args = get_arguments(argv)
    style = args['style']
    filePaths = args['files']
    if filePaths:
        if style:
            for filePath in filePaths:
                api.format_file(filePath=filePath, style=style, inPlace=args.get('in_place'))
        else:
            for filePath in filePaths:
                api.format_file(filePath=filePath, inPlace=args.get('in_place'))


def get_arguments(argv):
    '''
    Get arguments passed via command-line in dictionary.

    Paramters:
    argv: list
        List of arguments in sys.argv, including the first argument which is the script itself.
    
    Returns: dict
        A dictionary containing arguments for the formatter.
    '''
    parser = argparse.ArgumentParser(description='Formatter for SparkSQL queries.')

    parser.add_argument('-f', '--files', type=str, nargs='+', help='Paths to files to format.')

    parser.add_argument('-i', '--in-place', action='store_true', help='Format the files in place.')

    parser.add_argument('--style',
                        type=str,
                        default=None,
                        help="Style configurations for SparkSQL. Can be a path to a style config file or a dictionary.")

    args = vars(parser.parse_args(argv[1:]))

    return args


def run_main():
    '''
    Entry point for console_scripts in setup.py
    '''
    main(sys.argv)


if __name__ == '__main__':
    run_main()



import setuptools

with open('README.md', 'r') as fh:
    long_description = fh.read()

setuptools.setup(
    name='sparksqlformatter',
    version='0.1.5',
    author='largecats',
    author_email='linfanxiaolinda@outlook.com',
    description=
    'A SparkSQL formatter in Python based on https://github.com/zeroturnaround/sql-formatter, with customizations and extra features.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/largecats/sparksql-formatter',
    packages=setuptools.find_packages(),
    install_requires=['configparser'],
    classifiers=[
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    entry_points={
        'console_scripts': ['sparksqlformatter=sparksqlformatter:run_main'],
    })



# -*- coding: utf-8 -*-
# MIT License

# Copyright (c) 2016-present ZeroTurnaround LLC
# Copyright (c) 2016-present kufii
# Copyright (c) 2020-present largecats

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from __future__ import print_function  # for print() in Python 2
import re

from sparksqlformatter.src.tokenizer import TokenType, Tokenizer
from sparksqlformatter.src.indentation import Indentation
from sparksqlformatter.src.inline_block import InlineBlock
from sparksqlformatter.src.subquery import SubQuery
from sparksqlformatter.src.style import Style

trim_trailing_spaces = lambda s: re.sub(pattern='[ \t]+$', repl='', string=s)  # remove trailing spaces except \n


class Flag:
    '''
    Class for flags, or anotations, that can be added to tokens when formatting.
    '''
    SUBQUERY_ENDING_PAREN = 'SUBQUERY_ENDING_PAREN'
    INLINE = 'INLINE'  # don't add \n after comma in GROUP BY, ORDER BY, IN, etc


class Formatter:
    '''
    Class for formatting queries.
    '''
    def __init__(self, style=Style(), tokenOverride=None):
        '''
        Paramters
        style: sparksqlformatter.src.style.Style() object
            Styleurations for the query language.
        tokenOverride: function
            Function that takes token, previousKeyword and returns a token to overwrite given token (?).
        '''
        self.style = style
        self.indentation = Indentation(style.indent)
        self.inlineBlock = InlineBlock(style.inlineMaxLength)
        self.subQuery = SubQuery()
        self.tokenizer = Tokenizer(style=style)  # use the same styleurations as Formatter()
        self.tokenOverride = tokenOverride
        self.previousKeyword = None
        self.previousTopLevelKeyword = None
        self.tokens = []
        self.index = 0

    def format(self, query):
        '''
        Format query.
        
        Parameters
        query: str
            The query string.
        
        Return: str
            The formatted query.
        '''
        self.tokens = self.tokenizer.tokenize(input=query)  # identify tokens in the query
        formattedQuery = self.get_formatted_query_from_tokens()

        return formattedQuery.strip()

    def get_formatted_query_from_tokens(self):
        '''
        Create formatted query from identified tokens.

        Parameters

        Return: string
            The formatted query.
        '''
        formattedQuery = ''

        for i in range(len(self.tokens)):
            token = self.tokens[i]
            self.index = i

            if self.tokenOverride:
                token = self.tokenOverride(token, self.previousKeyword) or token

            if token.value.upper() == 'GROUP BY' or (token.value.upper() == 'ORDER BY'
                                                     and self.previousKeyword.value.upper() != 'PARTITION BY'):
                token.flag = Flag.INLINE

            if token.type == TokenType.WHITESPACE:
                # ignore
                continue
            elif token.type == TokenType.LINE_COMMENT:
                formattedQuery = self.format_line_comment(token, formattedQuery)
            elif token.type == TokenType.BLOCK_COMMENT:
                formattedQuery = self.format_block_comment(token, formattedQuery)
            elif token.type == TokenType.TOP_LEVEL_KEYWORD_NO_INDENT:
                formattedQuery = self.format_top_level_keyword_no_indent(token, formattedQuery)
                self.previousKeyword = token
            elif token.type == TokenType.TOP_LEVEL_KEYWORD:
                formattedQuery = self.format_top_level_keyword(token, formattedQuery)
                self.previousKeyword = token
                self.previousTopLevelKeyword = token
            elif token.type == TokenType.NEWLINE_KEYWORD:
                formattedQuery = self.format_newline_keyword(token, formattedQuery)
                self.previousKeyword = token
            elif token.type == TokenType.RESERVED_KEYWORD:
                formattedQuery = self.format_with_spaces(token, formattedQuery)
                self.previousKeyword = token
            elif token.type == TokenType.KEYWORD:
                formattedQuery = self.format_with_spaces(token, formattedQuery)
                self.previousKeyword = token
            elif token.type == TokenType.OPEN_PAREN:
                formattedQuery = self.format_opening_parentheses(token, formattedQuery)
            elif token.type == TokenType.CLOSE_PAREN:
                formattedQuery = self.format_closing_parentheses(token, formattedQuery)
            elif token.value == ',':
                formattedQuery = self.format_comma(token, formattedQuery)
            elif token.value == ':':
                formattedQuery = Formatter.format_without_spaces_before_with_space_after(token, formattedQuery)
            elif token.value == '.':  # no space before or after
                formattedQuery = Formatter.format_without_spaces(token, formattedQuery)
            elif token.value == ';':
                formattedQuery = self.format_query_separator(token, formattedQuery)
            elif token.value == '-':
                if i > 1:
                    offset = 1
                    while self.previous_token(offset=offset).type == TokenType.WHITESPACE:
                        offset += 1  # find most immediate previous token that is not white space
                    if self.previous_token(offset=offset).type in [
                            TokenType.KEYWORD, TokenType.RESERVED_KEYWORD, TokenType.NEWLINE_KEYWORD,
                            TokenType.TOP_LEVEL_KEYWORD, TokenType.TOP_LEVEL_KEYWORD_NO_INDENT
                    ]:
                        formattedQuery = Formatter.format_without_spaces_after(token, formattedQuery)
                    else:
                        formattedQuery = self.format_with_spaces(token, formattedQuery)
            else:
                formattedQuery = self.format_with_spaces(token, formattedQuery)

        return formattedQuery

    def format_line_comment(self, token, query):
        '''
        Format line comment.

        Parameters
        token: sparksqlformatter.src.token.Token() object
            Identified token of type TOKEN.LINE_COMMENT.
        query: string
            The query formatted so far.
        
        Return: string
            The query formatted so far together with the newly formatted line comment.
        '''
        return self.add_newline(query.rstrip() + ' ' + token.value)

    def format_block_comment(self, token, query):
        '''
        Format block comment.

        Parameters
        token: sparksqlformatter.src.token.Token() object
            Identified token of type Token.BLOCK_COMMENT.
        query: string
            The query formatted so far.
        
        Return: string
            The query formatted so far together with the newly formatted block comment.
        '''
        return self.add_newline(self.add_newline(query) + self.indent_comment(token.value))

    def indent_comment(self, comment):
        '''
        Indent comment.

        Parameters
        comment: string
            Value of the comment token.
        
        Return: string
            The comment with proper indentation.
        '''
        return re.sub(pattern='\n[ \t]*', repl=('\n' + self.indentation.get_indent()), string=comment)

    def format_top_level_keyword_no_indent(self, token, query):
        '''
        Format top-level keywords that are not indented.

        Paramaters
        token: sparksqlformatter.src.token.Token() object
            Identified token of type Token.TOP_LEVEL_KEYWORD_NO_INDENT.
        query: string
            The query formatted so far.
        
        Return: string
            The query formatted so far together with the newly formatted top-level keyword with no indent.
        '''
        self.indentation.decrease_top_level()
        query = self.add_newline(query) + Formatter.equalize_white_space(self.format_reserved_keyword(token.value))
        return self.add_newline(query)

    def format_top_level_keyword(self, token, query):
        '''
        Format top-level keywords with indentation.

        Parameters
        token: sparksqlformatter.src.token.Token() object
            Identified token of type Token.TOP_LEVEL_KEYWORD.
        query: string
            The query formatted so far.
        
        Return: string
            The query formatted so far together with the newly formatted top-level keyword.
        '''
        self.indentation.decrease_top_level()
        query = self.add_newline(query)
        self.indentation.increase_top_level()
        query += Formatter.equalize_white_space(self.format_reserved_keyword(token.value))
        return self.add_newline(query)

    def format_newline_keyword(self, token, query):
        if (self.previousKeyword.value.upper() in ['BETWEEN', 'WHEN'] and token.value.upper() in ['AND', 'OR']):
            return query + Formatter.equalize_white_space(self.format_reserved_keyword(token.value)) + ' '
        else:
            return self.add_newline(query) + Formatter.equalize_white_space(self.format_reserved_keyword(
                token.value)) + ' '

    @staticmethod
    def equalize_white_space(s):
        '''
        Replace any sequence of whitespace characters with single space.

        Parameters
        s: string
            The string to process.
        
        Return: string
            The string s with all sequences of whitespace characters turned to single space.
        '''
        return re.sub(pattern='\s+', repl=' ', string=s)

    def format_opening_parentheses(self, token, query):
        '''
        Opening parentheses increase the block indent level and start a new line.
        Take out the preceding space unless there was whitespace there in the original query or another opening parens or line comment.

        Parameters
        token: sparksqlformatter.src.token.Token() object
            Identified token of type Token.OPEN_PAREN.
        query: string
            The query formatted so far.
        
        Return: string
            The query formatted so far together with the newly formatted opening parentheses.
        '''
        preserveWhiteSpaceFor = [TokenType.WHITESPACE, TokenType.OPEN_PAREN, TokenType.LINE_COMMENT]
        if not (any(t == self.previous_token().type for t in preserveWhiteSpaceFor)):
            query = trim_trailing_spaces(query)
        query += token.value.upper() if self.style.reservedKeywordUppercase else token.value.lower()
        self.inlineBlock.begin_if_possible(self.tokens, self.index)
        if not self.inlineBlock.is_active():
            self.indentation.increase_block_level()
            query = self.add_newline(query)

        if self.previousKeyword.value.upper() == 'AS':  # start of subQuery, e.g., t0 AS (...)
            self.subQuery.started = True  # mark that subquery has started
            # This is to differentiate from opening/closng parentheses inside subquery
            # and to distinguish the starting opening parenthesis of the subquery
        self.subQuery.update(self, token)  # update subquery with the current token

        return query

    def format_closing_parentheses(self, token, query):
        '''
        Closing parentheses decrease the block indent level.

        Parameters
        token: sparksqlformatter.src.token.Token() object
            Identified token of type Token.CLOSE_PAREN.
        query: string
            The query formatted so far.
        
        Return: string
            The query formatted so far together with the newly formatted closing parentheses.
        '''
        token.value = token.value.upper() if self.style.reservedKeywordUppercase else token.value.lower()
        if (self.inlineBlock.is_active()):
            self.inlineBlock.end()
            query = Formatter.format_without_spaces_before_with_space_after(token, query)
        else:
            self.indentation.decrease_block_level()
            query = self.format_with_spaces(token, self.add_newline(query))

        self.subQuery.update(self, token)  # update subquery with the current token
        if self.subQuery.started and self.subQuery.matched():  # if this is the subquery's ending closing parenthesis
            token.flag = Flag.SUBQUERY_ENDING_PAREN  # add flag to mark this as subquery's ending parenthesis
            query = query.rstrip() + '\n' * (1 + self.style.linesBetweenQueries)  # add extra blank lines
            self.subQuery.reset()  # reset to start again
        return query

    def format_comma(self, token, query):
        '''
        Commas start a new line (unless within inline parentheses or SQL "LIMIT" clause).

        Parameters
        token: sparksqlformatter.src.token.Token() object
            Identified token with value ','.
        query: string
            The query formatted so far.
        
        Return: string
            The query formatted so far together with the newly formatted comma.
        '''
        # if this is the comma immediately after the closing parenthesis of a subquery, add extra blank line
        if self.previous_token().flag == Flag.SUBQUERY_ENDING_PAREN:
            query = query.strip()  # remove the \n added immediately after )
            self.subQuery.reset()
            return query + token.value + '\n' * (1 + self.style.linesBetweenQueries)  # add \n after ),
        query = trim_trailing_spaces(query) + token.value + ' '
        if (self.inlineBlock.is_active()):
            return query
        if re.search(pattern='^LIMIT$', string=self.previousKeyword.value):
            return query
        if self.previousTopLevelKeyword.flag == Flag.INLINE:
            if self.style.splitOnComma:
                return self.add_newline(query)
            else:
                inlineLength = len(query) - query.rfind('\n') + 1
                if inlineLength < self.style.inlineMaxLength:  # if fit in a line, don't split at comma
                    return query
                else:  # else, split at comma
                    indent = self.indentation.get_indent()
                    lastCommaIndex = query.rfind(',')
                    query = query[:lastCommaIndex + 1] + '\n' + indent + query[lastCommaIndex +
                                                                               2:len(query)]  # split at the last comma
                    return query
        else:
            return self.add_newline(query)

    @staticmethod
    def format_without_spaces_before_with_space_after(token, query):
        '''
        Add token to formatted query, removing spaces before token and adding space after it.

        Parameters
        token: sparksqlformatter.src.token.Token() object
            Current token to be formatted.
        query: string
            The query formatted so far.
        
        Return: string
            The query formatted so far together with the newly formatted token.
        '''
        return trim_trailing_spaces(query) + token.value + ' '

    @staticmethod
    def format_without_spaces(token, query):
        '''
        Add token to formatted query, removing spaces before token without adding space after it.

        Parameters
        token: sparksqlformatter.src.token.Token() object
            Current token to be formatted.
        query: string
            The query formatted so far.
        
        Return: string
            The query formatted so far together with the newly formatted token.
        '''
        return trim_trailing_spaces(query) + token.value

    @staticmethod
    def format_without_spaces_after(token, query):
        '''
        Add token to formatted query with no space after it.

        Parameters
        token: sparksqlformatter.src.token.Token() object
            Current token to be formatted.
        query: string
            The query formatted so far.
        
        Return: string
            The query formatted so far together with the newly formatted token.
        '''
        return query + token.value

    def format_with_spaces(self, token, query):
        '''
        Add token to formatted query with space after it.

        Parameters
        token: sparksqlformatter.src.token.Token() object
            Current token to be formatted.
        query: string
            The query formatted so far.
        
        Return: string
            The query formatted so far together with the newly formatted token.
        '''
        if token.type == TokenType.RESERVED_KEYWORD:
            value = self.format_reserved_keyword(token.value)
        elif token.type == TokenType.KEYWORD:
            value = self.format_keyword(token.value)
        else:
            value = token.value
        return query + value + ' '

    def format_reserved_keyword(self, value):
        '''
        Format reserved keyword, converting to uppercase of lowercase depending on self.style.reservedKeywordUppercase.

        Parameters
        value: string
            Value of the Token.RESERVED_KEYWORD token.
        query: string
            The query formatted so far.
        
        Return: string
            The query formatted so far together with the newly formatted reserved keyword.
        '''
        return value.upper() if self.style.reservedKeywordUppercase else value.lower()

    def format_keyword(self, value):
        '''
        Format keyword.

        Parameters
        value: string
            Value of the Token.KEYWORD token.
        query: string
            The query formatted so far.
        
        Return: string
            The query formatted so far together with the newly formatted keyword.
        '''
        # return value.upper() if self.style.reservedKeywordUppercase else value.lower()
        return value

    def format_query_separator(self, token, query):
        '''
        Format ';', which separates queries, including blank lines between queries.

        Parameters
        token: sparksqlformatter.src.token.Token() object
            Identified token with value ';'.
        query: string
            The query formatted so far.
        
        Return: string
            The query formatted so far together with the newly formatted ';'.
        '''
        self.indentation.reset_indentation()
        return trim_trailing_spaces(query) + token.value + '\n' * (self.style.linesBetweenQueries or 1)

    def add_newline(self, query):
        '''
        Add blank line after the end of query with proper indentation.

        Parameters
        query: string
            The query to process.
        
        Return: string
            The query with newly added blank line and indentation.
        '''
        query = trim_trailing_spaces(query)
        if not query.endswith('\n'):
            query += '\n'
        return query + self.indentation.get_indent()

    def previous_token(self, offset=1):
        '''
        Get previous token.

        Parameters
        offset: int
            The number of tokens to trace back.

        Return: sparksqlformatter.src.token.Token() object or None
            The token obtained by stepping backwards by given offset, if it exists.
            Return None if there is no such token.
        '''
        return self.tokens[self.index - offset] or None
